
import subprocess
import os
import sys
import shutil
from pathlib import Path


def init_checks():
    """
    Perform general checks,
    Check compatible python version,
    existence of 'config.local' file
    executable 'kubectl' command
    executable 'helm' command

    :return:
    """
    # Make sure python 3.5 and above is used
    if sys.version_info[0] < 3 or sys.version_info[1] < 5:
        print("Python 3.5 is a minimal requirement")
        exit(1)

    # Make sure config.local exists
    config_file = Path(str(Path.home()) + "/.kube/config.local")
    if not config_file.exists():
        print("Could not locate {}\nMake sure to create it".format(config_file))
        exit(1)

    os.environ["KUBECONFIG"] = str(config_file)

    # Make sure kubectl command exists
    if shutil.which("kubectl") is None:
        print("Could not execute 'kubectl' command\nMake sure 'kubectl' installed")
        exit(1)

    # .run return CompletedProcess with returncode and stdout and stderr values as bytes object, using UTF-8 decode
    cluster_info_output = subprocess.run(["kubectl", "cluster-info"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    # In case returncode is not 0
    if cluster_info_output.returncode:
        print(cluster_info_output.stderr.decode('utf-8') +
              "Make sure your cluster is up and running, and config file contains correct values")
        exit(1)
    # else:
    #     print(cluster_info_output.stdout.decode('utf-8'))

    # Make sure helm command exists
    if shutil.which("helm") is None:
        print("Could not execute 'helm' command,\n"
              "Make sure 'helm' installed: https://docs.helm.sh/using_helm/#installing-helm")
        exit(1)

    helm_init_output = subprocess.run(["helm", "init"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    # In case returncode is not 0
    if helm_init_output.returncode:
        print(helm_init_output.stderr.decode('utf-8'))
        exit(1)
    # else:
    #     print(helm_init_output.stdout.decode('utf-8'))




import subprocess


def get_cluster_info():
    cluster_info_completed_process = subprocess.run(["kubectl", "cluster-info"],
                                                    stdout=subprocess.PIPE,
                                                    stderr=subprocess.PIPE)

    if cluster_info_completed_process.returncode == 0:
        return cluster_info_completed_process.stdout.decode('utf-8').strip()
    else:
        return cluster_info_completed_process.stderr.decode('utf-8').strip()




import urwid
import urwid.raw_display
import urwid.web_display
from helm_functions import *
from init_checks import init_checks
from kubectl_functions import *
from helper import *


def main():

    text_header = (u"Helm charts installer -  "
                   u"UP / DOWN / PAGE UP / PAGE DOWN scroll.  F8 exits.")
    text_footer = remove_ansi_color_from_string(get_cluster_info().split("\n")[0])
    text_intro = [('important', u"Helm charts installer "),
                  u"installs kubernetes helm charts on a configured cluster, "
                  u"by default its intended to work on a cluster on the local host. "
                  u"\nPlease make sure to configure a .kube config file at ",
                  ('important', u"~/.kube/config.local")]

    text_repos_divider = [u"Installed ", ('important', u"Helm repositories:")]

    helm_repo_installed_repositories_text = urwid.Text(identify_installed_helm_repos(return_only_decoded_string=True).replace("\t", "    "))
    helm_repo_installed_repositories_dict = identify_installed_helm_repos()

    helm_repo_name_edit_box = urwid.Edit(('editcp', u"Helm repo name "), "set name", align='left', )
    helm_repo_url_edit_box = urwid.Edit(('editcp', u"Helm repo URL  "), "set url", align='left', )
    helm_repo_add_result = urwid.Text('')

    helm_charts_installed_dict = identify_installed_helm_charts()

    text_charts_divider = [u"Installed ", ('important', u"Helm charts:")]
    text_helm_charts_installed = identify_installed_helm_charts(return_only_decoded_string=True).replace("\t", "    ")

    text_button_list = [u"Yes", u"No", u"Perhaps", u"Certainly", u"Partially", u"Tuesdays Only", u"Help"]

    text_rb_list = [u"Morning", u"Afternoon", u"Evening", u"Weekend"]

    def button_press(button):
        frame.footer = urwid.AttrWrap(urwid.Text(
            [u"Pressed: ", button.get_label()]), 'header')

    def add_repo_button_pressed(button):
        """
        Execute once 'Add repository' button pressed,
        function will init the 'add_helm_repo' function,
        update the 'text_helm_repo_installed_repositories' urwid.TEXT object
        and update 'helm_repo_add_result' urwid.TEXT object

        :param button:
        :return:
        """

        add_helm_repo_output = add_helm_repo(helm_repo_name_edit_box.edit_text,
                                             helm_repo_url_edit_box.edit_text)['value']
        [urwid.AttrWrap(urwid.CheckBox("{:30}{}".format(txt['repo_name'], txt['repo_url'])), 'buttn', 'buttnf')
         for txt in identify_installed_helm_repos()]
        helm_repo_installed_repositories_text.set_text(identify_installed_helm_repos(return_only_decoded_string=True).replace("\t", "    "))
        helm_repo_add_result.set_text(add_helm_repo_output)

    radio_button_group = []

    blank = urwid.Divider()
    listbox_content = [
        blank,
        urwid.Padding(urwid.Text(text_intro), left=2, right=2, min_width=20),

        urwid.WidgetWrap(urwid.Divider("=", 1)),
        urwid.Padding(urwid.Text(text_charts_divider), left=2, right=2, min_width=20),
        urwid.WidgetWrap(urwid.Divider("*", 0, 1)),

        # Display helm charts block ###########
        # urwid.Padding(
        #     urwid.Pile([
        #         urwid.Text("    {:30}{}".format("NAME", "NAMESPACE")),
        #         urwid.Pile(
        #             [urwid.AttrWrap(urwid.CheckBox("{:30}{}".format(chart['chart_name'], chart['name_space'])), 'buttn',
        #                             'buttnf') for chart in helm_charts_installed_dict],
        #         ),
        #
        #     ]), left=2, right=2, min_width=10),
        #######################################
        # Helm charts display
        urwid.Padding(urwid.Text(text_helm_charts_installed), left=2, right=2, min_width=20),

        urwid.WidgetWrap(urwid.Divider("=", 1)),
        urwid.Padding(urwid.Text(text_repos_divider), left=2, right=2, min_width=20),
        urwid.WidgetWrap(urwid.Divider("*", 0, 1)),

        # Display helm repositories block #####
        # urwid.Padding(
        #     urwid.Pile([
        #         urwid.Text("    {:30}{}".format("NAME", "URL")),
        #         urwid.Pile(
        #             [urwid.AttrWrap(urwid.CheckBox("{:30}{}".format(repo['repo_name'], repo['repo_url'])), 'buttn',
        #                             'buttnf') for repo in helm_repo_installed_repositories_dict],
        #         ),
        #
        #     ]), left=2, right=2, min_width=10),
        #######################################

        # Helm repositories display
        urwid.Padding(helm_repo_installed_repositories_text, left=2, right=2, min_width=20),
        blank,
        urwid.Padding(urwid.AttrWrap(helm_repo_name_edit_box, 'editbx', 'editfc'), left=2, right=2),
        urwid.Padding(urwid.AttrWrap(helm_repo_url_edit_box, 'editbx', 'editfc'), left=2, right=2),
        blank,
        urwid.Padding(urwid.AttrWrap(urwid.Button("Add repository", add_repo_button_pressed), 'buttn', 'buttnf'), left=2, right=2, width=18),
        blank,
        urwid.Padding(helm_repo_add_result, left=2, right=2, min_width=20),
        blank,

        # urwid.WidgetWrap(urwid.Columns([
        #     urwid.Divider("#"),
        #     urwid.Divider('$'),
        #     urwid.Divider("&"),
        #     urwid.Divider('"'),
        #     urwid.Divider("'"),
        #     ])),
        #
        # blank,
        # urwid.Padding(urwid.GridFlow(
        #     [urwid.AttrWrap(urwid.Button(txt, button_press), 'buttn','buttnf') for txt in text_button_list],
        #     13, 3, 1, 'left'),
        #     left=4, right=3, min_width=13),
        #
        # blank,
        # urwid.Padding(urwid.GridFlow(
        #     [urwid.AttrWrap(urwid.RadioButton(radio_button_group, txt), 'buttn','buttnf')
        #         for txt in text_rb_list],
        #     13, 3, 1, 'left') ,
        #     left=4, right=3, min_width=13),
        # blank,
        ]

    header = urwid.AttrWrap(urwid.Text(text_header), 'header')
    footer = urwid.AttrWrap(urwid.Text(text_footer), 'footer')
    listbox = urwid.ListBox(urwid.SimpleListWalker(listbox_content))

    # Frame widget is used to keep the instructions at the top of the screen
    frame = urwid.Frame(urwid.AttrWrap(listbox, 'body'), header=header, footer=footer)

    palette = [
        ('body', 'black', 'light gray', 'standout'),
        ('reverse', 'light gray', 'black'),
        ('header', 'white', 'dark red', 'bold'),
        ('footer', 'white', 'dark green', 'bold'),
        ('important', 'dark blue', 'light gray', ('standout', 'underline')),
        ('editfc', 'white', 'dark blue', 'bold'),
        ('editbx', 'light gray', 'dark blue'),
        ('editcp', 'black', 'light gray', 'standout'),
        ('bright', 'dark gray', 'light gray', ('bold', 'standout')),
        ('buttn', 'black', 'light blue'),
        ('buttnf', 'white', 'dark blue', 'bold'),
        ]

    # use appropriate Screen class
    if urwid.web_display.is_web_request():
        screen = urwid.web_display.Screen()
    else:
        screen = urwid.raw_display.Screen()

    def unhandled(key):
        if key == 'f8':
            raise urwid.ExitMainLoop()

    urwid.MainLoop(frame, palette, screen, unhandled_input=unhandled).run()


def setup():
    urwid.web_display.set_preferences("Urwid Tour")
    # try to handle short web requests quickly
    if urwid.web_display.handle_short_request():
        return

    main()


if '__main__' == __name__ or urwid.web_display.is_web_request():
    print("Starting applications, please wait...")
    init_checks()
    setup()




import subprocess

supported_helm_deployments = [
    {'chart_name': 'ingress-traefik',
     'helm_repo_name': 'stable/traefik',
     'name_space': 'ingress-traefik',
     'values_file': 'values_files/ingress-traefik/values.local.yml'},

    {'chart_name': 'kubernetes-dashboard',
     'helm_repo_name': 'stable/kubernetes-dashboard',
     'name_space': 'kube-system',
     'values_file': 'values_files/kubernetes-dashboard/values.local.yml'},

    {'chart_name': 'nalkinscloud-nginx',
     'helm_repo_name': 'nalkinscloud/nalkinscloud-nginx',
     'name_space': 'nalkinscloud-nginx',
     'values_file': 'values_files/nalkinscloud-nginx/values.local.yml'},

    {'chart_name': 'jenkins',
     'helm_repo_name': 'stable/jenkins',
     'name_space': 'jenkins',
     'values_file': 'values_files/jenkins/values.local.yml'},
]


def identify_installed_helm_repos(return_only_decoded_string=False):
    """
    Function will perform a manipulation on a string output from the 'helm repo list' command
    If 'return_only_decoded_string' True, do not perform any string processing

    Returns an array of dicts with installed repos names and urls as strings
    as [{'repo_name': 'some_repo_name', 'repo_url': 'some_repo_url'}]

    'helm repo list' command return example:
    NAME        	URL
    stable      	https://kubernetes-charts.storage.googleapis.com
    local       	http://127.0.0.1:8879/charts
    nalkinscloud	https://arielevs.github.io/Kubernetes-Helm-Charts/

    by validating the first line, splitting by the tab delimiter,
    and checking that the first (0) value is 'NAME' and second (1) value is 'URL'
    an exception will be raised if the structure was change by HELM developers

    :param return_only_decoded_string: if True then return decoded (utf-8) "original" 'helm repo list' command output
    :return: array of dicts with repo installations
    """
    # Execute 'helm list' command, returned as CompletedProcess
    installed_repos_completed_process = subprocess.run(["helm", "repo", "list"],
                                                       stdout=subprocess.PIPE,
                                                       stderr=subprocess.PIPE)

    if return_only_decoded_string:
        if installed_repos_completed_process.returncode == 0:
            value = installed_repos_completed_process.stdout.decode('utf-8').strip()
        else:
            value = installed_repos_completed_process.stderr.decode('utf-8').strip()
        return value

    installed_repos = []

    # In case returncode is 0
    if not installed_repos_completed_process.returncode:
        # get stdout from installed_repos_completed_process, and decode for 'utf-8'
        # split stdout of installed_repos_completed_process by 'new line'
        installed_repos_stdout = installed_repos_completed_process.stdout.decode('utf-8').split("\n")

        # Perform validation on stdout of first (0) line
        first_line_stdout = installed_repos_stdout[0].split("\t")
        if first_line_stdout[0].strip() != 'NAME' or first_line_stdout[1].strip() != 'URL':
            raise Exception("'helm repo list' command output changed, "
                            "code change is needed to resolve this issue, "
                            "contact the developer.")

        # for every line in installed repos, excluding the headers line (NAME and URL)
        for line in installed_repos_stdout[1:]:
            # each stdout 'helm repo list' line composed by tabs delimiter, split it
            repo_details = line.split("\t")

            temp_dictionary = {}
            if repo_details[0] != "":
                # Add current line repo values to dict
                temp_dictionary.update({'repo_name': repo_details[0].strip()})
                temp_dictionary.update({'repo_url': repo_details[1].strip()})
                # Update final array with the temp array of dicts of current repo
                installed_repos.append(temp_dictionary)

    return installed_repos


def add_helm_repo(repo_name, repo_url):
    """
    Execute 'helm repo add' command on input values

    :param repo_name: name of repository as strings
    :param repo_url: url of repository as string
    :return: return code and value from execution command as dict
    """
    # execute and get CompletedProcess object
    completed_process_object = subprocess.run(["helm", "repo", "add", repo_name, repo_url],
                                              stdout=subprocess.PIPE,
                                              stderr=subprocess.PIPE)
    if completed_process_object.returncode == 0:
        value = completed_process_object.stdout.decode('utf-8')
    else:
        value = completed_process_object.stderr.decode('utf-8')
    return {'status': completed_process_object.returncode, 'value': value}


def remove_helm_repo(repos_array):
    """
    Execute 'helm repo remove' command on input values
    repos_array is an array of strings as:
    ['stable', 'local', 'nalkinscloud', ...]

    :param repos_array: array of strings
    :return: return code and value from execution command as dict
    """
    status = 0
    value = 'no errors found'
    for repo in repos_array:
        completed_process_object = subprocess.run(["helm", "repo", "remove", repo],
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE)
        # In case of a non 0 return code, update return from last iteration
        if completed_process_object.returncode != 0:
            status = completed_process_object.stderr.decode('utf-8')
            value = completed_process_object.stderr.decode('utf-8') + " *** Additional errors may occurred"

    return {'status': status, 'value': value}


def identify_installed_helm_charts(return_only_decoded_string=False):
    """
    Function will perform a manipulation on a string output from the 'helm list' command
    If 'return_only_decoded_string' True, do not perform any string processing

    Returns an array of dicts with installed chart names and namespaces as strings
    as [{'chart_name': 'some_chart_name', 'name_space': 'some_name_space'}]

    'helm list' command return example:
    NAME                	REVISION	UPDATED                 	STATUS  	CHART                     	NAMESPACE
    ingress-traefik     	2       	Thu Dec 27 19:45:01 2018	DEPLOYED	traefik-1.56.0            	ingress-traefik
    kubernetes-dashboard	11      	Sun Sep 16 11:21:24 2018	DEPLOYED	kubernetes-dashboard-0.7.3	kube-system

    by validating the first line, splitting by the tab delimiter,
    and checking that the first (0) value is 'NAME' and sixth (5) value is 'NAMESPACE'
    an exception will be raised if the structure was change by HELM developers

    :param return_only_decoded_string: if True then return decoded (utf-8) "original" 'helm list' command output
    :return: array of dicts with helm installations
    """
    # Execute 'helm list' command, returned as CompletedProcess
    installed_helm_completed_process = subprocess.run(["helm", "list"],
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.PIPE)

    if return_only_decoded_string:
        if installed_helm_completed_process.returncode == 0:
            value = installed_helm_completed_process.stdout.decode('utf-8').strip()
        else:
            value = installed_helm_completed_process.stderr.decode('utf-8').strip()
        return value

    installed_charts = []

    # In case returncode is 0
    if not installed_helm_completed_process.returncode:
        # get stdout from installed_helm_completed_process, and decode for 'utf-8'
        # split stdout of installed_helm_completed_process by 'new line'
        installed_helm_stdout = installed_helm_completed_process.stdout.decode('utf-8').split("\n")

        # Perform validation on stdout of first (0) line
        first_line_stdout = installed_helm_stdout[0].split("\t")
        if first_line_stdout[0].strip() != 'NAME' or first_line_stdout[5].strip() != 'NAMESPACE':
            raise Exception("'helm list' command output changed, "
                            "code change is needed to resolve this issue, "
                            "contact the developer.")

        # for every line in installed charts, excluding the headers line (Name, Revision, Updated etc...)
        for line in installed_helm_stdout[1:]:
            # each stdout 'helm list' line composed by tabs delimiter, split it
            chart_details = line.split("\t")

            temp_dictionary = {}
            if chart_details[0] != "":
                # Add current line chart values to dict
                temp_dictionary.update({'chart_name': chart_details[0].strip()})
                temp_dictionary.update({'name_space': chart_details[5].strip()})
                # Update final array with the temp array of dicts of current helm deployment
                installed_charts.append(temp_dictionary)

    return installed_charts


def install_helm_charts(charts_array):
    """
    Execute 'helm update' command on input values
    charts_array is an array of strings as:
    ['ingress-traefik', 'kubernetes-dashboard', ...]

    installation process will iterate over the 'supported_helm_deployments' variable,
    which contains an array of dicts with all supported helm charts deployments,
    installation process will init only if 'charts_array' has a value that also exists is 'supported_helm_deployments'

    :param charts_array: array of strings
    :return: return code and value from execution command as dict
    """
    # Update helm repo before installation
    subprocess.run(["helm", "repo", "update"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    status = 0
    value = 'no errors found'
    for deployment in supported_helm_deployments:
        # Only if current chart name exists in the input 'charts_array'
        if deployment['chart_name'] in charts_array:
            completed_process_object = subprocess.run(["helm", "upgrade", deployment['chart_name'],
                                                       "--install", deployment['helm_repo_name'],
                                                       "--namespace", deployment['name_space'],
                                                       "-f", deployment['values_file']],
                                                      stdout=subprocess.PIPE,
                                                      stderr=subprocess.PIPE)
            # In case of a non 0 return code, update return from last iteration
            if completed_process_object.returncode != 0:
                status = completed_process_object.stderr.decode('utf-8')
                value = completed_process_object.stderr.decode('utf-8') + " *** Additional errors may occurred"
    return {'status': status, 'value': value}


def delete_helm_installations(charts_array):
    """
    Execute 'helm delete' command on input values
    charts_array is an array of strings as:
    ['ingress-traefik', 'kubernetes-dashboard', ...]

    :param charts_array: array of strings
    :return: return code and value from execution command as dict
    """
    status = 0
    value = 'no errors found'
    for installation in charts_array:
        completed_process_object = subprocess.run(["helm", "delete", "--purge", installation],
                                                  stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE)
        # In case of a non 0 return code, update return from last iteration
        if completed_process_object.returncode != 0:
            status = completed_process_object.stderr.decode('utf-8')
            value = completed_process_object.stderr.decode('utf-8') + " *** Additional errors may occurred"
    return {'status': status, 'value': value}
